# coding: utf-8
"""
Function

----------------------------------
Version    : 0.0.1
Date       : 2023/3/31   18:25
----------------------------------
Author     : April
Contact    : fanglwh@outlook.com
"""

__all__ = ["EqualizedFocalLoss"]

from typing import Optional

import torch
from torch import Tensor

from .loss import GeneralizedCrossEntropyLoss


class EqualizedFocalLoss(GeneralizedCrossEntropyLoss):
    """

    Args:
        num_classes: The number of classes. Default: ``1000``.
        focal_gamma:  . Default: ``2.0``.
        focal_alpha:  . Default: ``0.25``.
        scale_factor:  . Default: ``8.0``.
        fpn_levels:  . Default: ``5``.
        name: Name of the loss function. Default: ``equalized_focal_loss``.
        reduction: Reduction type, choice of mean, none, sum. Default: ``none``.
        loss_weight: Loss weight.  Default: ``1.0``.
        ignore_index: The class index to be ignored. Default: ``-1``.
    """
    def __init__(self,
                 num_classes: int = 1000,
                 focal_gamma: float = 2.0,
                 focal_alpha: float = 0.25,
                 scale_factor: float = 8.0,
                 fpn_levels: int = 5,
                 name: str = "equalized_focal_loss",
                 reduction: str = "mean",
                 loss_weight: float = 1.0,
                 ignore_index: int = -1,
                 ) -> None:
        activation_type = "sigmoid"
        GeneralizedCrossEntropyLoss.__init__(self,
                                             name=name,
                                             reduction=reduction,
                                             loss_weight=loss_weight,
                                             activation_type=activation_type,
                                             ignore_index=ignore_index)

        # cfg for focal loss
        self.focal_gamma = focal_gamma
        self.focal_alpha = focal_alpha

        # ignore bg class and ignore idx
        self.num_classes = num_classes - 1

        # cfg for efl loss
        self.scale_factor = scale_factor
        # initial variables
        self.register_buffer("pos_grad", torch.zeros(self.num_classes))
        self.register_buffer("neg_grad", torch.zeros(self.num_classes))
        self.register_buffer("pos_neg", torch.ones(self.num_classes))
        self.cache_mask = None
        self.cache_target = None

        # grad collect
        self.grad_buffer = []
        self.fpn_levels = fpn_levels

    def collect_grad(self, grad_in):
        bs = grad_in.shape[0]
        self.grad_buffer.append(grad_in.detach().permute(0, 2, 3, 1).reshape(bs, -1, self.num_classes))
        if len(self.grad_buffer) == self.fpn_levels:
            target = self.cache_target[self.cache_mask]
            grad = torch.cat(self.grad_buffer[::-1], dim=1).reshape(-1, self.num_classes)

            grad = torch.abs(grad)[self.cache_mask]
            pos_grad = torch.sum(grad * target, dim=0)
            neg_grad = torch.sum(grad * (1 - target), dim=0)

            # allreduce(pos_grad)
            # allreduce(neg_grad)

            self.pos_grad += pos_grad
            self.neg_grad += neg_grad
            self.pos_neg = torch.clamp(self.pos_grad / (self.neg_grad + 1e-10), min=0, max=1)

            self.grad_buffer = []

    def forward(self,input: Tensor, target: Tensor, normalizer: Optional[float] = None, **kwargs) -> Tensor:
        n_c = input.shape[-1]
        input = input.reshape(-1, n_c)
        target = target.reshape(-1)
        n_i, _ = input.size()

        expand_target = input.new_zeros(n_i, n_c + 1)
        expand_target[torch.arange(n_i), target] = 1
        expand_target = expand_target[:, 1:]

        sample_mask = (target != self.ignore_index)

        inputs = input[sample_mask]
        targets = expand_target[sample_mask]
        self.cache_mask = sample_mask
        self.cache_target = expand_target

        pred = torch.sigmoid(inputs)
        pred_t = pred * targets + (1 - pred) * (1 - targets)

        map_val = 1 - self.pos_neg.detach()
        dy_gamma = self.focal_gamma + self.scale_factor * map_val
        # focusing factor
        ff = dy_gamma.view(1, -1).expand(n_i, n_c)[sample_mask]
        # weighting factor
        wf = ff / self.focal_gamma

        # ce_loss
        ce_loss = -torch.log(pred_t)
        cls_loss = ce_loss * torch.pow((1 - pred_t), ff.detach()) * wf.detach()

        # to avoid an OOM error
        # torch.cuda.empty_cache()

        if self.focal_alpha >= 0:
            alpha_t = self.focal_alpha * targets + (1 - self.focal_alpha) * (1 - targets)
            cls_loss = alpha_t * cls_loss

        if normalizer is None:
            normalizer = 1.0

        return self._reduce(cls_loss, normalizer=normalizer)
